"""
Helpers and such for smallwp.
"""

import bz2
import cPickle
import os

from smallwp import wpprefs

def makeNamespaceChecker(lang, format):
	"""returns a function taking a title and returning true if the title either
	is non-namespaced or has a namespace we want to keep.
	"""
	keptNamespaces = wpprefs.getLangData(lang, "keptNamespaces")
	if format=="XML":
		keptNamespaces = keptNamespaces | wpprefs.getLangData(
			lang, "xmlKeptNamespaces")
	elif format=="HTML":
		keptNamespaces = keptNamespaces | set(
			[wpprefs.getLangData(lang, "categoryNamespace")])
	def isInterestingNamespace(pageTitle):
		if ":" in pageTitle:
			namespace, _ = pageTitle.split(":", 1)
			return namespace in keptNamespaces
		return True
	return isInterestingNamespace


class AccessorContainer(object):
	def createRwAttribute(self, name, initVal):
		setattr(self, name, initVal)
		setattr(self, "set"+name.capitalize(),
			lambda val: setattr(self, name, val))
		setattr(self, "get"+name.capitalize(),
			lambda: getattr(self, name))


class Article(AccessorContainer):
	def __init__(self):
		self.createRwAttribute("title", None)
		self.createRwAttribute("text", None)


class ArticleWriter(object):
	def __init__(self, targetDir, opts, targetSize=3e6, encoding="utf-8"):
		self.targetDir = targetDir
		self.opts = opts
		self.targetSize = targetSize
		self.encoding = encoding
		self.curSerial = 0
		self.curFile = None
		self._prepareTarget()
		self.pageIndexName = os.path.join(targetDir, "pages.index")
		self.pageIndex = open(self.pageIndexName, "w")
		self.openNextFile()
	
	def _prepareTarget(self):
		if not os.path.exists(self.targetDir):
			os.makedirs(self.targetDir)
		for fName in os.listdir(self.targetDir):
			os.unlink(os.path.join(self.targetDir, fName))

	def openNextFile(self):
		if self.curFile is not None:
			self.curFile.close()
		self.curSerial += 1
		self.curFileName = "%07d.wpd"%self.curSerial
		self.curFile = bz2.BZ2File(os.path.join(self.targetDir,
			self.curFileName), "w")
		self.curSize = 0

	def writeArticle(self, article):
		content = article.getText()
		if isinstance(content, unicode):
			content = content.encode(self.encoding,
				"xmlcharrefreplace")
		if self.curSize+len(content)>self.targetSize:
			self.openNextFile()
		self.pageIndex.write("%s\t%s\t%d\t%d\n"%(
			article.getTitle().encode("utf-8").replace("\t", " "),
			self.curFileName, self.curSize, len(content)))
		self.curFile.write(content)
		self.curSize = self.curFile.tell()

	def _fixIndex(self):
		self.pageIndex.close()
# The next line would be bad if targetDir was writable to anyone else.
		tmp = os.path.join(self.targetDir, "tmp%d"%os.getpid())
		os.system("LC_ALL=C sort %s > %s"%(self.pageIndexName, tmp))
		os.rename(tmp, self.pageIndexName)

	def _writeMeta(self):
		metaDict = {
			"lang": self.opts.lang,
			"isHTML": self.opts.inputFormat=="HTML",
		}
		f = open(wpprefs.getMetaLocation(), "w")
		cPickle.dump(metaDict, f)
		f.close()

	def finalize(self):
		self.curFile.close()
		self.curFile = None
		self.curFileName = None
		self._fixIndex()
		self._writeMeta()
		cPickle.dump(
			makeSections(self.pageIndexName, 40),
			open(os.path.join(self.targetDir, "sections.pickle"), "w"))


def makeSections(pagesFileName, targetLength):

	def buildHierarchy(sectionsList, targetLength):
		if len(sectionsList)<targetLength:
			return sectionsList
		thisLevelSections = []
		startIndex = 0
		while 1:
			endIndex = startIndex+targetLength
			if endIndex>len(sectionsList)-10:
				thisLevelSections.append((sectionsList[startIndex][0],
					sectionsList[-1][1], tuple(sectionsList[startIndex:])))
				break
			else:
				thisLevelSections.append((sectionsList[startIndex][0],
					sectionsList[endIndex-1][1], 
					tuple(sectionsList[startIndex:endIndex])))
			startIndex += targetLength
		return buildHierarchy(thisLevelSections, targetLength)

	def getKey(indexLine):
		return indexLine.decode("utf-8").split("\t")[0]

	sections = []
	f = open(pagesFileName)
	startKey, startPos = getKey(f.readline()), 0
	curNumItems = 1
	while 1:
		pos = f.tell()
		curLine = f.readline()
		if not curLine:
			sections.append(((startKey, startPos), 
				(getKey(curLine), pos), ()))
			break
		if curNumItems>=targetLength:
			sections.append(((startKey, startPos), 
				(getKey(lastLine), pos), ()))
			startKey, startPos = getKey(curLine), pos
			curNumItems = 0
		lastLine = curLine
		curNumItems += 1
	return buildHierarchy(sections, targetLength)

