"""
Code to generate and store math images.

TODO: expire old images.
"""

import md5
import os
import glob
import tempfile
import threading

import wpprefs

latexTemplate = r"""
\documentclass{article}
\usepackage{amsmath}
\usepackage{amsfonts}
\usepackage{amssymb}
\pagestyle{empty}
\newcommand{\N}{\mathbb{N}}
\newcommand{\R}{\mathbb{R}}
\newcommand{\Z}{\mathbb{Z}}
\begin{document}
%s
\end{document}
"""


class Error(Exception):
	pass


def texToPngs(teXSource, filesToLinkBackTo=[]):
	"""returns a list of png, with each png corresponding to one page of teXSource.
	"""
	def readImages(pattern):
		fNames = glob.glob(pattern)
		fNames.sort()
		images = []
		for fName in fNames:
			f = open(fName)
			images.append(f.read())
			f.close()
		return images

	owd = os.getcwd()
	wdName = tempfile.mkdtemp("texToPs", "tmp", "/tmp")
	try:
		os.chdir(wdName)
		for fname in filesToLinkBackTo:
			os.symlink(os.path.join(owd, fname), fname)
		f = open("tmp.tex", "w")
		f.write("\\batchmode\n%s\n"%teXSource.encode("iso-8859-1"))
		f.close()
		os.system("latex tmp.tex > /dev/null")
		#open("/home/msdemlei/bla.tex", "w").write(open("tmp.tex").read())
		os.system("dvipng -Ttight -D %d -q -o 'img%%05d.png' tmp.dvi"%
			wpprefs.getPref("mathOutputResolution"))
		imgs = readImages("img*.png")
		log = open("tmp.log").read()
	finally:
		os.chdir(owd)
		os.system("rm -rf '%s'"%wdName)
	return imgs


class MathImages:
	"""is a container for math images.

	Do not access _buildQueue directly but only using _withBuildQueue.
	It is conceivable that you'll get races otherwise.
	"""
	def __init__(self):
		self._buildQueue = []
		self.buildQueueLock = threading.Lock()
		self.pngs = {}

	def _withBuildQueue(self, callable, *args, **kwargs):
		"""calls callable with the buildqueue as the first argument
		and args, kwargs as further arguments.

		It first acquires a lock on the buildqueue and will block until
		that lock is released.
		"""
		self.buildQueueLock.acquire()
		try:
			return callable(self._buildQueue, *args, **kwargs)
		finally:
			self.buildQueueLock.release()

	def _getAndClearBuildQueue(self):
		"""returns the current content of the build queue and clears it.

		That means that you should better do something with what this
		method returns.
		"""
		def getAndClear(buildQueue):
			values = buildQueue[:]
			buildQueue[:] = []
			return values
		return self._withBuildQueue(getAndClear)

	def _enqueue(self, key, texCode):
		self.pngs[key] = None
		self._withBuildQueue(lambda q, s: q.append(s), (key, texCode))

	def getImageKey(self, texCode):
		"""queues up texCode for generation if necessary and returns a handle
		(string) that can later be used to retrieve the corresponding png.
		"""
		key = md5.new(texCode.encode("utf-8")).hexdigest()
		if not self.pngs.has_key(key):
			self._enqueue(key, texCode)
		return key

	def _buildSource(self, mathParts):
		"""returns TeX source that hopefully builds each item of mathParts
		on a page of its own.
		"""
		return latexTemplate%("\\clearpage\n".join(
			["%s"%part for part in mathParts]))

	def _buildQueuedMath(self):
		"""builds images for math snippets build queue and updates the
		image dictionary accordingly.

		This is called automatically by getImage if necessary.

		This could be a race if someone would enter something into the
		buildqueue -- if this thing got multithreaded (god forbid!),
		we'd have to use locking.
		"""
		buildQueue = self._getAndClearBuildQueue()
		if not buildQueue:
			return
		for index, png in enumerate(
				texToPngs(self._buildSource([part[1] for part in buildQueue]))):
			self.pngs[buildQueue[index][0]] = png

	def getImage(self, key):
		if not self.pngs.has_key(key):
			raise Error("Math item %s unknown\n"%key)
		if self.pngs[key]==None:
			self._buildQueuedMath()
		return self.pngs[key]

_mathImages = MathImages()

def getImageKey(texCode):
	return _mathImages.getImageKey(texCode)

def getImage(key):
	return _mathImages.getImage(key)
