"""
Split and recompress the huge HTML-only dumps of wikipedia available
at http://static.wikipedia.org
"""

import re
import os
import sys
import traceback

from smallwp import common
from smallwp import wpprefs

## The following code lifted in part from tarfile_scanner by
# Uwe Schmidt, see http://www.procoders.net/wp-content/tarfile_scanner.zip
# stripped down for the current use case.

BLOCKSIZE = 512

def trim_null(v):
    return v[:v.find('\000')]

def oct2int(v):
    return int(v, 8)

def int2int(v):
    v = v.strip('\00 ')
    if v:
        return int(v)
    return None    

class TarInfo:
    def __init__(self, hblock):
        self.name     = trim_null(hblock[0:100])
        self.size     = oct2int(hblock[124:136])
        blocks, remainder = divmod(self.size, BLOCKSIZE)
        if remainder:
            blocks += 1
        self.size_in_block = blocks*BLOCKSIZE


def iterTarFile(fp):
	ct = 0
	while 1:
		hblock = fp.read(BLOCKSIZE)
		if not ord(hblock[0]): 
			break
		header = TarInfo(hblock)
		header.fileIndex = ct
		size = header.size_in_block
		content = fp.read(size)
		yield header, content[:header.size]
		ct += 1

## End code inspired by Uwe Schmidt


def iter7zTarFile(source):
	return iterTarFile(os.popen("7za x -so %s"%source))


def manglePath(path):
	"""returns our internal representation of WP paths, i.e., all leading
	path components and the extension .html stripped.
	"""
	return path.split("/")[-1][:-5]


def filterArticles(tarIter, lang):
	boringPattern = wpprefs.getLangData(lang, "htmlIgnoredPat")
	for tarInfo, content in tarIter:
		name = os.path.basename(tarInfo.name)
		if not name.endswith(".html") or boringPattern.match(name):
			continue
		if name.startswith("@"):
			continue
		yield tarInfo, content


class PathTranslations(object):
	"""hold the information what names map to which titles.

	As a side effect, it also does redirects.  The main trick here is that
	we believe file names are unique within a dump and thus strip all
	path information from both paths and destinations.
	"""
	def __init__(self, destName):
		self.destName = destName
		self.dest = open(self.destName, "w")

	def add(self, path, destination):
		path, destination = manglePath(path), destination.encode("utf-8")
		if path==destination:
			return
		self.dest.write("%s\t>%s\n"%(path, destination))

	def addRedirect(self, path, destination):
		self.dest.write(
			"%s\t%s\n"%(manglePath(path), manglePath(destination)))

	def finalize(self):
		self.dest.close()
# The next line would be bad if targetDir was writable to anyone else.
		tmp = self.destName+"%d"%os.getpid()
		os.system("LC_ALL=C sort %s > %s"%(self.destName, tmp))
		os.rename(tmp, self.destName)


class Article(common.Article):
# RE-based manipulations to save time...
	def __init__(self, rawText):
		common.Article.__init__(self)
		self._parse(rawText)

	startMarker = "<!-- start content -->"
	endMarker = "<!-- end content -->"
	def _extractContent(self, rawText):
		startPos = rawText.find(self.startMarker)
		endPos = rawText.find(self.endMarker)
		return rawText[startPos+len(self.startMarker):endPos]

	dirtyLinkPat = re.compile(r'href="\.[^"]*/([^/"]*.html)"')
	def _cleanupLinks(self, content):
		return self.dirtyLinkPat.sub(r'href="/q/\1"', content)

	editSectionPat = re.compile('<span class="editsection">.*?</span>')
	def _nukeEditLinks(self, content):
		return self.editSectionPat.sub('', content)

	commentPat = re.compile("(?s)<!-- .*? -->") # This may cost
		# more CPU than it's worth the savings, but let's do it anyway.
	def _killComments(self, content):
		return self.commentPat.sub("", content)

	titlePat = re.compile('<h1 class="firstHeading">(.*?)</h1>')
	def _parse(self, rawText):
		self.setTitle(self.titlePat.search(rawText).group(1).decode("utf-8"))
		content = self._extractContent(rawText)
		content = self._cleanupLinks(content)
		content = self._nukeEditLinks(content)
		content = self._killComments(content)
		self.setText(content)


redirectPat = re.compile('http-equiv="Refresh" content="0;url=([^"]*.html)"')

def isRedirect(content):
	mat = redirectPat.search(content, 0, 1000)
	if mat:
		return mat.group(1)


def splitSource(src, opts):
	destdir = wpprefs.getPref("basedir")
	writer = common.ArticleWriter(destdir, opts)
	trans = PathTranslations(os.path.join(destdir, "translations.txt"))
	isInterestingNamespace = common.makeNamespaceChecker(opts.lang, 
		opts.inputFormat)
	nextThreshold = 100
	for info, content in filterArticles(iter7zTarFile(src), opts.lang):
		try:
#			if info.fileIndex>20000:
#				break
			r = isRedirect(content)
			if r:
				trans.addRedirect(info.name, r)
			else:
				art = Article(content)
				if not isInterestingNamespace(art.getTitle()):
					continue
				trans.add(info.name, art.getTitle())
				writer.writeArticle(art)
			if info.fileIndex>nextThreshold:
				sys.stderr.write("\r%8d..."%info.fileIndex)
				sys.stderr.flush()
				nextThreshold += 100
		except:
			#traceback.print_exc()
			print "Don't know what this is:", info.name
			#open("zw.html", "w").write(content)
	writer.finalize()
	trans.finalize()
