"""
Splits wikipedia contents into manageable blocks.
"""

import bz2
import cPickle
import os
import re
import sys
import textwrap
import xml
from xml.sax.handler import ContentHandler

from smallwp import common
from smallwp import splithtmldump
from smallwp import wpprefs


class StartStopHandler(ContentHandler):
	def __init__(self):
		ContentHandler.__init__(self)
		self.currentChars = []
		self.attStack = []

	def startElement(self, name, attrs):
		self.attStack.append(attrs)
		if hasattr(self, "_start_"+name):
			getattr(self, "_start_"+name)(name, attrs)
		self.clearCharContent()
	
	def endElement(self, name):
		attrs = self.attStack.pop()
		if hasattr(self, "_end_"+name):
			getattr(self, "_end_"+name)(name, attrs)
		self.clearCharContent()

	def characters(self, chars):
		self.currentChars.append(chars)
	
	def getCharContent(self):
		stuff = u"".join(self.currentChars)
		return stuff
	
	def clearCharContent(self):
		self.currentChars = []


class ArticleReader(StartStopHandler):
	def __init__(self, articleWriter, opts):
		StartStopHandler.__init__(self)
		sys.stderr.write("Splitting source, written articles:\n")
		self.articleWriter = articleWriter
		self._checkNamespace = common.makeNamespaceChecker(opts.lang, 
			opts.inputFormat)
		self.pageCounter = 0

	def setDocumentLocator(self, locator):
		self.getCurrentLine = locator.getLineNumber

	def _start_page(self, name, attrs):
		self.curArticle = common.Article()

	def _end_page(self, name, attrs):
		if (":" in self.curArticle.getTitle() and 
				not self._checkNamespace(self.curArticle.getTitle())):
			return
		self.articleWriter.writeArticle(self.curArticle)
		self.pageCounter += 1
		if not self.pageCounter%100:
			sys.stderr.write("%8d\r"%self.pageCounter)
#		if self.pageCounter>2000: raise IOError("Enough")

	def _end_title(self, name, attrs):
		self.curArticle.setTitle(self.getCharContent())

	def _end_text(self, name, attrs):
		self.curArticle.setText(self.getCharContent())


def splitSource(inputFile, opts):
	destdir = wpprefs.getPref("basedir")
	writer = common.ArticleWriter(destdir, opts)
	try:
		xml.sax.parse(inputFile, ArticleReader(writer, opts))
	finally:
		writer.finalize()
	try:
		os.unlink(os.path.join(destdir, "isHTML"))
	except os.error: # we're not overwriting an HTML split, so don't bother
		pass


_langPats = map(re.compile, [
# regexps for file names helping _guessLang.  The first one matching wins.
	r"(\w+)wiki-\d+-pages-articles",
	r"wikipedia-(\w+)-html.*",
])

def _guessLang(sourceName):
	"""returns a language code guessed from sourceName.

	It knows the structure of wikipedia dump names as of early 2009.
	If it cannot guess the code, it raises a ReportableError.
	"""
	for pat in _langPats:
		mat = pat.match(sourceName)
		if mat:
			return mat.group(1)
	raise ReportableError("Cannot guess language from '%s'.  Use -l option."%
		sourceName)


def parseCommandLine():
	from optparse import OptionParser
	parser = OptionParser(usage="%%prog [options] <dump file>"
		" -- imports a wikipedia dump file.  Both HTML and XML versions"
		" are supported and autodetected using the input file name.")
	parser.add_option("-f", "--force-format", help="override autodetection"
		" of input format; the argument is either XML or HTML.", action="store",
		type="str", default=None, dest="inputFormat")
	parser.add_option("-l", "--lang", help="Use language preferences for"
		" language LANG.", type="str", default=None, dest="lang", action="store")
	parser.add_option("-H", "--help-config", help="Dump help on options in"
		" ~/.smallwp and exit.", dest="helpConfig", action="store_true",
		default=False)

	opts, args = parser.parse_args()
	if opts.helpConfig:
		wpprefs.showHelp()
		sys.exit(0)
	if len(args)!=1:
		parser.print_help()
		sys.exit(1)
	if opts.inputFormat is None:
		if "xml" in args[0]:
			opts.inputFormat = "XML"
		elif "html" in args[0]:
			opts.inputFormat = "HTML"
		else:
			raise wpprefs.ReportableError(
				"Could not autodetect input format.  Use -f flag.")
	if opts.lang is None:
		opts.lang = _guessLang(args[0])
	return opts, args


def main():
	try:
		opts, args = parseCommandLine()
		if opts.inputFormat=="XML":
			splitSource(bz2.BZ2File(args[0]), opts)
		elif opts.inputFormat=="HTML":
			splithtmldump.splitSource(args[0], opts)
		else:
			raise wpprefs.ReportableError("Unknown input format: %s"%opts.inputFormat)
	except wpprefs.ReportableError, msg:
		sys.stderr.write(textwrap.fill(unicode(msg).encode("utf-8"))+"\n")
