# -*- coding: iso-8859-1 -*-
"""
Code to handle templates.

While the template syntax is a relatively sane part of the wikimedia syntax,
it's still quite convoluted, and some attempts do write a well-performing
pyparsing-based grammar failed.  So, I hand-craft a somewhat hacky parser
here.
"""

import weakref
import sys
import traceback
import time
import re
import textwrap

import wpprefs

class TemplateError(Exception):
	pass


class Cache(object):
	def __init__(self):
		self.data = {}
	
	def store(self, name, item):
		self.data[name] = (item, time.gmtime())
	
	def get(self, name):
		return self.data.get(name)[0]
	
	def has(self, name):
		return self.data.has_key(name)

_treecache = Cache()


openingLit = re.compile(r"{{{|{{#|{{|\[\[|<noinclude>|<includeonly>")
closingLit = re.compile(r"(}}}|}}|\]\]|</noinclude>|</includeonly>)")
typeMap = {
	"{{{": "tVar", 
	"{{#": "tFunc", 
	"{{": "tpl", 
	"[[": "wl", 
	"<noinclude>": "noinc", 
	"<includeonly>": "inconly",
}

closerMap = {
	"tVar": "}}}",
	"tFunc": "}}",
	"tpl": "}}",
	"wl": "]]",
	"noinc": "</noinclude>",
	"inconly": "</includeonly>",
}


def getType(lit):
	return typeMap[lit]

def getCloser(el):
	return closerMap[el]


def removeComments(txt):
	parts, cPos = [], 0
	while True:
		nPos = txt.find("<!--", cPos)
		if nPos==-1:
			break
		parts.append(txt[cPos:nPos])
		cPos = txt.find("-->", nPos)+3
		assert cPos!=-1
	parts.append(txt[cPos:])
	return "".join(parts)


class Node(object):
	outputWidth = 80
	def __init__(self, type, initchildren=None):
		self.type = type
		if initchildren:
			self.children = initchildren[:]
		else:
			self.children = []

	def append(self, item):
		if item=='':
			return
		self.children.append(item)

	def prepend(self, item):
		self.children.insert(0, item)

	def __str__(self, indent=""):
		res = [indent+"<<%s>>"%self.type]
		indent = indent+" "
		for child in self.children:
			if isinstance(child, basestring):
				res.append("\n".join(textwrap.wrap(child.strip(), 
					width=self.outputWidth,
					initial_indent=indent, subsequent_indent=indent)))
			else:
				res.append(child.__str__(indent))
		return "\n".join(res)
	
	def __repr__(self):
		return "<%s Node %s>"%(self.type, 
			", ".join(repr(c) for c in self.children))


class ArgumentNode(Node):
	"""is a node keeping elements of argument lists.

	The reason it gets a specialized class is that we want the built-in
	capability to tell named arguments from others.

	On the first call to isNamed, the Node is split into nameChildren
	and contentChildren if it is a named node.  Later calls do not reparse.
	"""
	def __init__(self, initchildren=None):
		Node.__init__(self, "arg")
		if initchildren:
			for c in initchildren:
				self.append(c)
		self._isNamed = None

	def __repr__(self):
		if self._isNamed is None:
			return "<unparsed arg %s>"%repr(self.children)
		else:
			if self._isNamed:
				return "<arg %s %s>"%(repr(self.name), repr(self.children))
			else:
				return "<arg %s>"%repr(self.children)

	def isNamed(self):
		"""returns true if this is a named argument.

		Unfortunately, the syntax for named arguments is not entirely clear.
		If I assume a grammar like [A-Za-z]+\s*=, the pain is that I
		may have to expand substantial parts of the argument.

		I simplify the rules for now: I take all nodes that are strings up
		to the first non-string node.  This means that names can't be the
		result of a template expansion.
		"""
		if self._isNamed is None:
			self._isNamed = False
			leadingStrings = []
			for child in self.children:
				if not isinstance(child, basestring):
					break
				leadingStrings.append(child)
			maybeName = "".join(leadingStrings)
			mat = re.match('''(?L)\s*([^!"#$%&'()*+,./+:;<=>?@^`{|}~\s]+)'''
				"\s*=(.*)", maybeName)
			if mat:
				self.name = mat.group(1)
				self.children = [mat.group(2)]+self.children[
					len(leadingStrings):]
				self._isNamed = True
		return self._isNamed


TOPLEVEL="<toplevel>"

def _buildTree(txt, tName, validate=True):
	"""does lexical analysis and parse tree building for txt.

	Lexing the template syntax is a complete nightmare due to ambiguities
	of all kinds.  E.g., {{{!}} should be lexed as <literal {> <template !>, which
	really is only possible after building the tree.  Greedy lexing
	doesn't work.  Everything else isn't attractive either.  Sigh.

	All this stuff should probably be done in a tricky recursive descent parser.
	"""
	parts, cPos = [], 0
	nodeStack = [Node("root")]
	omat, cmat = openingLit.search(txt, cPos), closingLit.search(txt, cPos)
	while True:
		if not omat and not cmat:
			break
		if omat and omat.start()<cmat.start():
			if cPos!=omat.start():
				nodeStack[-1].append(txt[cPos:omat.start()])
			nodeStack.append(Node(getType(omat.group())))
			cPos = omat.end()
			omat = openingLit.search(txt, cPos)
		else:
			if cPos!=cmat.start():
				nodeStack[-1].append(txt[cPos:cmat.start()])
			newChild = nodeStack.pop()
			expectedCloser = getCloser(newChild.type)
			if expectedCloser!=cmat.group().strip():
				# Fix things like {{x{{ab}}}}
				if cmat.group().startswith(expectedCloser):
					closerRe = re.compile(re.escape(expectedCloser))
					cmat = closerRe.match(txt, cmat.start())
				#Fix {{{!}}  (Yikes!)
				elif expectedCloser.endswith(cmat.group()):
					nodeStack[-1].append("{")
					newChild.type = "tpl"
				elif validate:
					raise TemplateError("Mismatched closer near %s, expected %s"
						" (char %d, %s)"%(txt[cmat.start():cmat.start()+40], 
						getCloser(newChild.type), cmat.start(), tName.strip()))
			nodeStack[-1].append(newChild)
			cPos = cmat.end()
			cmat = closingLit.search(txt, cPos)
	while len(nodeStack)>1:
		print "Warning: Unclosed %s node while parsing %s"%(
			nodeStack[-1].type, tName.strip())
		newChild = nodeStack.pop()
		nodeStack[-1].append(newChild)
	if txt[cPos:].strip():
		nodeStack[0].append(txt[cPos:])
	return nodeStack[0]


def getTreeFromDisk(tName, data, validate=False):
	tName = tName.strip()
	if not _treecache.has(tName):
		txt = removeComments(data.getArticle(tName))
		getTree(txt, tName, validate)
	return _treecache.get(tName)

def getTree(txt, tName=TOPLEVEL, validate=False):
	tName = tName.strip()
	if tName==TOPLEVEL:
		return _buildTree(txt, tName, validate)
	if not _treecache.has(tName):
		_treecache.store(tName, _buildTree(txt, tName, validate))
	return _treecache.get(tName)


class Arguments(list):
	"""represents the arguments of a template.

	You can either access the arguments by number (and the number may even
	be is a string) or by name (if named arguments are used).  We count
	arguments as wikimedia does, but $0 is the template name itself.

	This thing must do quite a bit of bookkeeping, so only add arguments
	using the append method.

	The main point of this class is to avoid unnecessary expansions, so
	we keep arguments unexpanded as long as possible.
	"""
	default = ""

	def __init__(self, context):
		super(Arguments, self).__init__()
		self.context = context
		self.namedArgs = {}
		self.argCache = {}

	def _computeItem(self, index):
		try:
			numericIndex = int(index)
			if numericIndex<len(self):
				return self._getExpandedArg(
					super(Arguments, self).__getitem__(numericIndex))
			else:
				return self.default
		except ValueError:
			if self.namedArgs.has_key(index):
				return self[self.namedArgs[index]]
			else:
				return self.default

	def __getitem__(self, index):
		if not self.argCache.has_key(index):
			self.argCache[index] = self._computeItem(index)
		return self.argCache[index]

	def get(self, key, default):
		preDefault = self.default
# Ouch.  We should think of sth better.
		self.default = default
		val = self[key]
		self.default = preDefault
		return val

	def asDict(self):
		res = {}
		for index in range(len(self)):
			res[str(index)] = self[index]
		for name, index in self.namedArgs.iteritems():
			res[name] = res.get(str(index), "")
		return res

	def append(self, arg):
		if self and arg.isNamed():  # 0th arg (name) is never named
			name = arg.name
			self.namedArgs[name] = len(self)
		super(Arguments, self).append(arg)

	def _getExpandedArg(self, arg):
		res = makeWikitext(self.context, arg).strip()
		return res

	_argSep = re.compile(r"(^|[^\\])\|")

	def _splitArgString(self, str):
		cPos = 0
		parts = []
		while True:
			mat = self._argSep.search(str, cPos)
			if not mat:
				break
			parts.append(str[cPos:mat.start()]+mat.group(1))
			cPos = mat.end()
		parts.append(str[cPos:])
		return parts

	@classmethod
	def fromNode(cls, context, node):
		instance, curArg = cls(context), ArgumentNode()
		for child in node.children:
			if isinstance(child, basestring):
				argStrings = instance._splitArgString(child)
				curArg.append(argStrings.pop(0))
				for argString in argStrings:
					instance.append(curArg)
					curArg = ArgumentNode([argString])
			else:
				curArg.append(child)
		instance.append(curArg)
		return instance


def tplDebug(name, args):
	print ">>>>>>>>>>>", name, args
	return ""

def tplAbs(name, args):
	return str(abs(float(args[1])))


def _makeServerPrefix():
	if wpprefs.getPref("address"):
		return "http://%s:%s"%(wpprefs.getPref("address"),
			wpprefs.getPref("port")),
	else:
		return "/"


# XXX TODO: We'll need to figure out how to get at much of this info.
_predefinedTemplates = {
	"SITENAME": wpprefs.getPref("sitename"),
	"SERVER": _makeServerPrefix(),
	"SERVERNAME": wpprefs.getPref("address"),
	"SCRIPTPATH": "",
	"CURRENTVERSION": "Aw, whatever",
	"CONTENTLANGUAGE": "Some language",

	"NAMESPACE": "",
	"NAMESPACEE": "",
	"SUBJECTSPAGE": "",
	"SUBJECTSPAGEE": "",
	"SUBJECTSPAGENAME": "",
	"SUBJECTSPAGENAMEE": "",
	"PAGENAME": "",
	"PAGENAMEE": "",
	"TALKSPACE": "",
	"TALKSPACEE": "",
	"TALKPAGENAME": "",
	"TALKPAGENAMEE": "",
	"FULLPAGENAME": "DIESE SEITE",
	"FULLPAGENAMEE": "DIESE SEITE/U",
	"BASEPAGENAME": "BASISSEITE",
	"BASEPAGENAMEE": "BASISSEITE/U",
	"REVISIONID": "0",
	"REVISIONDAY": "0",
	"REVISIONDAY2": "00",
	"REVISIONMONTH": "00",
	"REVISIONYEAR": "2007",
	"REVISIONTIMESTAMP": "000000000000",
	"SUBPAGENAME": "DIESE UNTERSETE",
	"SUBPAGENAMEE": "DIESE UNTERSEITE/U",
	"CURRENTMONTH": lambda *args: time.strftime("%m"),
	"LOCALMONTH": lambda *args: time.strftime("%m"),
	"CURRENTMONTHNAME": lambda *args: time.strftime("%B"),
	"LOCALMONTHNAME": lambda *args: time.strftime("%B"),
	"CURRENTMONTHNAMEGEN": lambda *args: time.strftime("%Bs"),
	"LOCALMONTHNAMEGEN": lambda *args: time.strftime("%Bs"),
	"CURRENTMONTHABBREV": lambda *args: time.strftime("%b"),
	"LOCALMONTHABBREV":  lambda *args: time.strftime("%b"),
	"CURRENTDAY": lambda *args: time.strftime("%d"),
	"LOCALDAY": lambda *args: time.strftime("%d"),
	"CURRENTDAY2": lambda *args: time.strftime("%d"),
	"LOCALDAY2": lambda *args: time.strftime("%d"),
	"CURRENTDOW": lambda *args: time.strftime("%w"),
	"CURRENTDAYNAME": lambda *args: time.strftime("%A"),
	"LOCALDAYNAME":  lambda *args: time.strftime("%A"),
	"CURRENTWEEK": "",  # XXX implement
	"LOCALWEEK": "",
	"CURRENTYEAR": "",
	"LOCALYEAR": "",
	"CURRENTHOUR": "",
	"LOCALHOUR": "",
	"CURRENTTIME": "",
	"LOCALTIME": "",
	"CURRENTTIMESTAMP": "",
	"LOCALTIMESTAMP": "",

	"NUMBEROFEDITS": "0",
	"NUMBEROFPAGES": "0", # XXX
	"NUMBEROFARTICLES": "0", # XXX
	"NUMBEROFFILES": "",
	"NUMBEROFUSERS": "0",
	"NUMBEROFADMINS": "0",

# The following template(s) may be informally defined -- I haven't found
# them documented, just used :-)
	"abs": tplAbs,
	"subst": tplDebug,
}

_magicNamespaces = {
	"DEFAULTSORT": "",
	"DISPLAYTITLE": "",
	"INT": "",
	"PLURAL": "",
	"anchorencode": "",
	"filepath": "",
	"formatnum": lambda name, tArgs: str(name), 
	"lc": "",
	"lcfirst": "",
	"padleft": "",
	"padright": "",
	"urlencode": "",
	"ns": "",
	"uc": "",
	"localurl": "",
}


class WikitextContext(object):
	def __init__(self, data, vars={}, env={}):
		self.vars = vars
		self.env = env
		self.data = data

	@classmethod
	def fromArgs(cls, parent, args):
		parentVars = args.asDict()
		return cls(parent.data, parentVars, parent.env)


def _joinChildren(context, node):
	"""expands all children of node and returns them all joined together.
	"""
	res = []
	for child in node.children:
		if isinstance(child, basestring):
			res.append(child)
		else:
			res.append(makeWikitext(context, child))
	return "".join(res)


def _splitNode(context, node, splitter):
	"""returns a pair (aString, rest) of the node content split by splitter.

	The rest is another Node or an empty string.  This is used for switch.
	"""
	pre, children = [], node.children[:]
	while children:
		child = children.pop(0)
		childS = makeWikitext(context, child)
		if splitter in childS:
			preS, postS = childS.split(splitter, 1)
			pre.append(preS)
			break
		else:
			postS = None
			pre.append(childS)
	if children or postS:
		return "".join(pre), Node(node.type, [postS]+children)
	else:
		return "".join(pre), ""


def make_tVar(context, node):
	args = Arguments.fromNode(context, node)
	val = context.vars.get(args[0], args[1])
	return val


def tFunc_special(context, args):
	# This we don't need, I guess
	return ""


def tFunc_language(context, args):
#XXX do we need this?
	return ""


_leadingBlanksPat = re.compile("\n +")
def unindent(aString):
# Yet another hack: it seems wikimedia undoes indents in at least 
# if parameters.  Oh, my.
	return _leadingBlanksPat.sub("\n", aString)

def tFunc_if(context, args):
	cond = args[0]
	if cond.strip():
		return unindent(args[1])
	else:
		return unindent(args[2])
	return ""


def tFunc_ifexist(context, args):
	lemma = args[0]
	if context.data.has_key(lemma):
		return args[1]
	else:
		return args[2]


def tFunc_ifeq(context, args):
	val1, val2 = args[0], args[1]
	if val1.strip()==val2.strip():
		return args[2]
	else:
		return args[3]


def tFunc_expr(context, args):
# XXX TODO: do real parsing -- this way, it's a gaping security hole
	return "0"
	return str(eval(args[0]))


def tFunc_ifexpr(context, args):
# XXX TODO: as for _tFunc_expr
	return args[1]
	if eval(args[0]):
		return args[1]
	else:
		return args[2]


def tFunc_switch(context, args):
	def parseCases(args):
		default = ""
		caseDict = {}
		curHeads = []
		for arg in args:
			head, tail = _splitNode(context, arg, "=")
			head = head.strip()
			if tail:
				curHeads.append(head)
				for head in curHeads:
					if head=="#default" or head=='':
						default = tail
					else:
						caseDict[head] = tail
				curHeads = []
			else:
				curHeads.append(head)
		return caseDict, default
	caseDict, default = parseCases(args[1:])
	value = args[0]
	return makeWikitext(context, caseDict.get(value, default)).strip()


def tFunc_time(context, args):
	# XXX TODO: implement this
	if len(args)==2:
		return args[1]
	else:
		return time.asctime()


def tFunc_titleparts(context, args):
	numParts = int(args[1])
	if len(args)>2:
		startIndex = int(args[2])-1
	else:
		endIndex = 0
	if numParts==0:
		return args[0]
	else:
		return args[0].strip().split("/")[startIndex:startIndex+numParts]


_tFuncRE = re.compile("(?s)\s*(\w+)\s*:(.*)")
def make_tFunc(context, node):
# XXX todo: move tree manipulations to parser.
	if hasattr(node, "func"):
		funcName = node.func
	else:
		mat = _tFuncRE.match(node.children[0])
		funcName, node.children[0]  = mat.groups()
		node.func = funcName
	return globals()["tFunc_"+funcName](context,
		Arguments.fromNode(context, node))


_tplSpecials = {
	"!":  "|",
	"!-": "|-",
	"!!": "||",
	"!-!":"|-|",
	"!+": "|+",
	"!~": "|-!",
	"(!": "{|",
	"!)": "|}",
}


def _loadTemplate(context, fullName, tName, tArgs):
	if "_" in fullName:  # XXXX What's the truth on underscores in names?
		fullName = fullName.replace("_", " ")
	context = WikitextContext.fromArgs(context, tArgs)
	return makeWikitext(context, getTreeFromDisk(fullName, context.data))


def _getTemplateMaker(context, ns, name, tArgs):
	"""returns something to get a replacement text for the template
	name with parsed arguments tArgs.

	This can either be a string that is directly inserted, or a
	callable taking the namespaceless name, the argument dictionary,
	and the current environment that returns an insertable string.
	"""
	if name in _tplSpecials:
		return _tplSpecials[name]
	if name in _predefinedTemplates:
		return _predefinedTemplates[name]
	if ns in _magicNamespaces:
		return _magicNamespaces[ns]
	return lambda *args: _loadTemplate(context, ns+":"+name,
		name, tArgs)

def make_tpl(context, node):
	tArgs = Arguments.fromNode(context, node)
	tName = tArgs[0]
	try:
		ns, name = tName.split(":", 1)
	except ValueError:
		ns = wpprefs.getLangData(context.data.lang, "templateNamespace")
		name = tName
	try:
		expander = _getTemplateMaker(context, ns, name, tArgs)
		if isinstance(expander, basestring):
			return expander
		else:
			return expander(name, tArgs)
	except wpprefs.PageNotFound, msg:
		print "Missing template %s"%msg.title.encode("utf-8")
		return "(Missing template %s)"%msg.title
	except:
		traceback.print_exc()
		return "(Bad template %s, see log)"%tName

def make_noinc(context, node):
	return ""

def make_inconly(context, node):
	return _joinChildren(context, node)

def make_wl(context, node):
	# Wikilinks are only handled because the WP parser seems to ignore
	# bars in them.
	content = _joinChildren(context, node)
	return "[[%s]]"%content

def make_root(context, node):
	return _joinChildren(context, node)

def make_arg(context, node):
	return _joinChildren(context, node)


def makeWikitext(context, node):
	if isinstance(node, basestring):
		res = node
	else:
		res = globals()["make_"+node.type](context, node)
	if not isinstance(res, basestring):
		raise Exception("%s didn't return a string on node %s"%("make_"+node.type,
			repr(node)))
	return res


def formatTemplate(templateLiteral, data):
	context = WikitextContext(data)
	return makeWikitext(context,
		getTree(removeComments(templateLiteral)))


def main():
	print unicode(formatTemplate(open(sys.argv[1]).read().decode("utf-8"))).encode("utf-8")

if __name__=="__main__":
	context = WikitextContext(vars={"Adresse": "foo",
		"Art": "Stadt"})
	print repr(makeWikitext(context, 
	getTree("""
{{#switch: {{{Art|}}}
	| <!-- leer -->
	| Ortsgemeinde = Gemeinde
	| Amt = Amts
	| Gemeindeverwaltungsverband
	| Samtgemeinde
	| Verbandsgemeinde
	| Verwaltungsgemeinschaft
	| Verwaltungsverband = Verbands
	| #default = {{{Art}}}
}}verwaltung:""", validate=True)))
	#main()
