# -*- coding: utf-8 -*-
"""
An attept to do a compact and sane converter for wikitext to html that
does 80% right and doesn't badly mess up the rest.

This is again mainly RE mess.  But there's little I can do about this --
a real grammar of wikitext doesn't exist and is *extremely* hairy to come
up with.
"""

import re
import urllib
import random
import sys
import traceback

from smallwp import mathimages
from smallwp import wpprefs
from smallwp import templates

class Redirect(Exception):
	pass


jsFunctions = """
<script language="javascript">
function blowUp(destId) {
	element = document.getElementById(destId);
	element.style.visibility='visible';
	element.style.height='auto';
}

function collapse(destId) {
	element = document.getElementById(destId);
	element.style.visibility='hidden';
	element.style.height='0px';
}

</script>
"""


def getRandomId():
	return str(random.randint(0, 1073741824))


class Table:
	"""is a container for a wiki table.
	"""
	def __init__(self, attrs):
		self.attrs = attrs
		self.rows = []
		self.caption = None
		self.startNewRow()

	def addCaption(self, caption):
		self.caption = caption

	def startNewRow(self, attrs=""):
		self.curRow = []
		self.rows.append((attrs, self.curRow))

	def _addToLastCell(self, additionalContent):
		try:
			type, attrs, content = self.curRow[-1]
			self.curRow[-1] = type, attrs, content+"\n"+additionalContent
		except IndexError:  #Junk at start of table
			self.curRow.append(("td", "", additionalContent))

	def _parseCellDesc(self, cellDesc, type):
		if "|" in cellDesc:
			attrs, content = cellDesc.split("|")
		else:
			attrs, content = "", cellDesc
		self.curRow.append((type, attrs, content))

	def addCell(self, cellLine):
		if cellLine.startswith("*") or cellLine.startswith("<"):
			self._addToLastCell(cellLine)
			return
		if cellLine.startswith("|") or cellLine.startswith("!"): 
			cellSplitter = cellLine[0]*2
			if cellSplitter=="!!":
				type = "th"
			else:
				type = "td"
			for cellDesc in cellLine[1:].split(cellSplitter):
				try:
					self._parseCellDesc(cellDesc, type)
				except ValueError:
					print "Bad table line:", cellDesc.encode("utf-8")
		else:
			# Yikes.  This isn't in the spec.
			self._addToLastCell(cellLine)

	def _formatCell(self, type, attrs, content):
		return "<%s %s>%s</%s>"%(type, attrs, content, type)

	def _formatRow(self, attrs, cells):
		if cells:
			return "<tr %s>%s</tr>"%(attrs,
				"\n".join([self._formatCell(*cell) for cell in cells]))
		else:
			return ""
		
	def asHtml(self):
		capStr = ""
		if self.caption:
			capStr = "<caption>%s</caption"%self.caption
		return "<table %s>%s%s</table>"%(self.attrs, capStr, "\n".join(
			[self._formatRow(*row) for row in self.rows]))


# XXX TODO Do something with this junk
_magicWords = {
	"__TOC__": "",
	"__NOTOC__": "",
	"__FORCETOC__": "",
	"__NOEDITSECTION__": "",
	"__NEWSECTIONLINK__": "",
	"ISBN": "",
	"PMID": "",
	"RFC": "",
}

class WikitextFrobnicator(object):
	"""is a converter from wiki markup to html.

	This can be reused, but it uses instance storage and is thus
	*not* thread-safe.
	"""
	def __init__(self, wpBase, data):
		self.wpBase = wpBase
		self.data = data
		self.processors = [
			self.handleRedirects,
			self.handleComments,
			self.handleMath,
			self.handleReferences,
			self.handleTemplates,
#			self.handleHTMLEscaping,
			self.handleApos, 
			self.handleLinks, 
			self.handleTables,
			self.handleHeadlines,
			self.handleEnums,
			self.handleParagraphs,]
		self.knownNamespaces = {
			wpprefs.getLangData(self.data.lang, "categoryNamespace"): 
				"_formatCategory",
		}

	blocklevelElements = set(["h1", "h2", "h3", "h4", "h5", "div",
		"table", "tr", "td", "th", "ul", "ol", "li"])

	def _insertHidden(self, someHtml):
		"""adds someHtml for later inclusion in the document and returns a
		wiki-neutral proxy for it.

		This is used to keep finished markup from further wiki formatting.
		"""
# XXX maybe use a random marker here
		key = "/Hidden/%d/HIDDEN/"%(len(self.hiddenParts))
		self.hiddenParts[key] = someHtml
		return key

	def _unhideHidden(self, txt):
		return re.sub("/Hidden/[0-9]+/HIDDEN/", 
			lambda mat: self.hiddenParts[mat.group(0)], txt)

	def _startsWithBlocklevel(self, ln):
		"""returns true if ln starts with an HTML blocklevel element.

		This is used in various places in the later stages of processing
		to decide what to do -- e.g., whether to insert p-tags.
		"""
		if ln.startswith("|"):
			return True
		mat = re.match(r"</?(\w+)", ln)
		if not mat:
			return False
		return mat.group(1) in self.blocklevelElements

	def _parseRecursiveElement(self, opener, closer, elHandler, txt):
		"""processes possibly recursive elements delimited by opener and
		closer using elHandler.

		elHandler receives the element content as a string argument.

		>>> w = WikitextFrobnicator(None, None)
		>>> import sys
		>>> def hdlr(stuff): return "<noo>%s</noo>"%stuff
		>>> w._parseRecursiveElement("{{", "}}", hdlr, "ab {{foo}} cd")
		"ab <noo>foo</noo> cd"
		>>> w._parseRecursiveElement("{{", "}}", hdlr, "ab {{foo||{{bar}}"
		...   "|{{no}}}} cd")
		'ab <noo>foo||<noo>bar</noo>|<noo>no</noo></noo> cd'
		"""
		tokenStack = [[]]
		for element in re.split("(%s|%s)"%(re.escape(opener), 
				re.escape(closer)), txt):
			if element==opener:
				tokenStack.append([])
			elif element==closer:
				if len(tokenStack)>1:
					res = elHandler("".join(tokenStack.pop()))
					tokenStack[-1].append(res)
				else:
					self.wikitextSyntaxError("Spurious closer %s ignored"%closer)
			else:
				tokenStack[-1].append(element)
		if len(tokenStack)>1:
			self.wikitextSyntaxError("Missing closer %s. Ignoring corresponding"
				" opener(s)."%closer)
			while len(tokenStack)>1:
				stuff = "".join(tokenStack.pop())
				tokenStack[-1].append(stuff)
		return "".join(tokenStack[0])

	def wikitextSyntaxError(self, msg):
		""" "handles" an error in wikitext.

		This *may* raise an error or do nothing at all.  Calling functions
		should assume it returns and at least do *something* to recover.
		"""
		sys.stderr.write(msg+"\n")

	def _formatTable(self, tBody):
		lines = tBody.split("\n")
		table = Table(lines[0])
		for line in lines[1:]:
			if not line.strip():
				pass
			elif line.startswith("|+"):
				table.addCaption(line[2:])
			elif line.startswith("|-"):
				table.startNewRow(line[2:])
			else:
				table.addCell(line)
		return table.asHtml()

	def handleTables(self, txt):
		# In theory, we should require some special characters in front
		# of each table line, like this:
		# re.sub(r"(?m)^(\{\|.*\n(?:^[ *|!<-].*\n|^\n)*?^\|\}.*)", 
		# In practice, this just doesn't work
		return self._parseRecursiveElement("{|", "|}", self._formatTable, txt)

	def handleComments(self, txt):
		return re.sub("<!--.*?-->", "", txt)

	def handleHTMLEscaping(self, txt):
		return txt.replace("&", "&amp;").replace("<", "&lt;")

	def _formatTemplatePlain(self, templateContent):
		items = templateContent.split("|")
		templateClass = items[0].strip()
		arguments = [item.strip() for item in items[1:]]
		return ('<div class="templateHead"'
			' onMouseover="blowUp(\'%(id)s\')"'
			' onMouseout="collapse(\'%(id)s\')">'
			'<em>Template: %(head)s</em></div>'
			'<div class="templateBody" style="visibility:hidden;height:0px"'
			' id="%(id)s">'
			'%(body)s</div>')%{
				"head": templateClass, 
				"body": " -- ".join(arguments),
				"id": getRandomId()}

	def _formatTemplateFancy(self, templateContent):
		"""will be the next generation template code.

		It actually can include templates, but what's in them is
		such a bad mess that we keep this shut off for now.
		"""
		try:
			return templates.formatTemplate("{{"+templateContent+"}}", self.data)
		except:
			traceback.print_exc()
			return "(Bad template %s, see log)"%templateContent

	if wpprefs.getPref("doTemplates"):
		_formatTemplate = _formatTemplateFancy
	else:
		_formatTemplate = _formatTemplatePlain

	def handleTemplates(self, txt):
		templateStuff = self._parseRecursiveElement(
			"{{", "}}", self._formatTemplate, txt)
#		try:
#			open("afterTemplates.txt", "w").write(templateStuff.encode("utf-8"))
#		except IOError:
#			pass
		return templateStuff

	def _handleReference(self, refContent):
		self.references.append(refContent)
		ind = len(self.references)
		return self._insertHidden(
			' <a name="refback_%03d" href="#ref_%03d">[%d]</a>'%(ind, ind, ind))

	def handleReferences(self, txt):
		self.references = []
		return self._parseRecursiveElement("<ref>", "</ref>", 
			self._handleReference, txt)
		

	def handleParagraphs(self, txt):
		"""assumption: block level elements start on a line of their own.
		"""
		inBlockLevel = True
		lines = []
		for line in txt.split("\n"):
			if inBlockLevel:
				if not self._startsWithBlocklevel(line):
					lines.append("<p>")
					inBlockLevel = False
			else:
				if self._startsWithBlocklevel(line):
					lines.append("</p>")
					inBlockLevel = True
				if not line.strip():
					line = "</p><p>"
			lines.append(line)
		return "\n".join(lines)

		return "<p>%s</p>"%("</p>\n<p>".join(re.split("\n\s*\n", txt)))

	def handleApos(self, txt):
		txt = re.sub("(?s)'''''(.*?)'''''", r"<strong><em>\1</em></strong>", txt)
		txt = re.sub("(?s)'''(.*?)'''", r"<strong>\1</strong>", txt)
		return re.sub("(?s)''(.*?)''", r"<em>\1</em>", txt)
	
	def handleHeadlines(self, txt):
		def mkHx(matob):
			return "<h%d>%s</h%d>"%(len(matob.group(1)), matob.group(2), 
				len(matob.group(1)))
		return re.sub(r"(?s)(==+)(.*?)\1", mkHx, txt)

	enumChars = "[*#:]"

	def _getEnumChanges(self, prev, cur):
		"""returns (deletions, additions, prefix)

		prev and cur have a (possibly empty) common prefix.  After that,
		everything in prev is a deletion, and everything in cur is an
		additon.
		"""
		def getCommonPrefix(s1, s2):
			index = 0
			for index, (a, b) in enumerate(zip(s1, s2)):
				if a!=b:
					break
			return s1[:index]
		prefix = getCommonPrefix(prev, cur)
		return prev[len(prefix):], cur[len(prefix):], prefix

	enumCharToHtml = {
		"*": "ul",
		"#": "ol",
		":": 'ul style="list-style-type:none"'}

	def _formatEnumChanges(self, prevOpener, curOpener):
		"""returns the necessary opening and closing tags to get from
		wikitext enumeration markup to html enumeration markup.
		"""
		result = []
		closers, openers, prefix = self._getEnumChanges(prevOpener, curOpener)
		for closer in closers:
			result.append("</%s>"%self.enumCharToHtml[closer].split()[0])
		for opener in openers:
			result.append("<%s>"%self.enumCharToHtml[opener])
		return "\n".join(result)

	def _formatEnumRange(self, matob):
		result = []
		lastOpener = ""
		openerRe = re.compile("%s+"%self.enumChars)
		for ln in matob.group(0).split("\n"):
			if not ln.strip():
				continue
			curOpener = openerRe.match(ln).group()
			if lastOpener!=curOpener:
				result.append(self._formatEnumChanges(lastOpener, curOpener))
				lastOpener = curOpener
			result.append("<li>%s</li>"%ln[len(curOpener):].strip())
		result.append(self._formatEnumChanges(curOpener, ""))
		return "\n".join(result)+"\n"

	def handleEnums(self, txt):
		enumRe = "(?m)(^%s+.*\n+)+"%self.enumChars
		return re.sub(enumRe, self._formatEnumRange, txt)

	def _collectMath(self, matob):
		isDisplay = not not matob.group(1)
		mathFrag = matob.group(2)
		dispClass = {True: "mathdisplay", False: "math"}[isDisplay]
		tag = {True: "div", False: "span"}[isDisplay]
		left, right = {True: (r"\begin{equation}", r"\end{equation}"),
			False: (r"\(", r"\)")}[isDisplay]
		mathFrag = "%s%s%s"%(left, mathFrag, right)
		return self._insertHidden(
			' <%s class="%s"><img src="/math/%s" alt="%s"></%s> '%(
			tag, dispClass, mathimages.getImageKey(mathFrag), mathFrag, tag))

	def handleMath(self, txt):
		mathRe = re.compile("(?s)(:?) ?<math>(.*?)</math>")
		return mathRe.sub(self._collectMath, txt)

	def _formatLink(self, url, title):
		return '<a href="%s">%s</a>'%(url, title)

	def _formatFreeLink(self, linkContent):
		if " " in linkContent:
			url, content = linkContent.split(" ", 1)
		else:
			url, content = linkContent, linkContent
		if not re.match("\w+://", url): # Hack to not destroy the page when no URL
			return linkContent            # ends up from, e.g., templates
		return self._formatLink(url, content)

	def _formatCategory(self, wWord):
		return '<span style="category">%s</span>'%wWord

	def _formatUnknownNamespace(self, wWord):
		return '<span class="unresolved" title="%s">o</span>'%wWord

	def _makeNamespaceLink(self, wWord):
		if ":" in wWord:
			namespace, target = wWord.split(":", 1)
			return getattr(self, self.knownNamespaces.get(namespace,
				"_formatUnknownNamespace"))(wWord)

	def _computeWikiLink(self, wWord):
		return "%s%s"%(self.wpBase, 
			urllib.quote_plus(wWord.encode("utf-8")))

	def _formatWikiLink(self, linkContent):
		hint = ""
		if "|" in linkContent:
			wWord, title = linkContent.split("|", 1)
		else:
			wWord, title = linkContent, linkContent
		if ":" in wWord:
			return self._makeNamespaceLink(wWord)
		return self._formatLink(self._computeWikiLink(wWord), title)

	def _formatNakedLinks(self, txt, 
			urlRE=re.compile(r"(.)(http://[^\s<>\"']+)")):
		"""inserts markup for links not yet in html elements.

		This is quick and ugly, let's fix it as we go along (there should be
		lots of useful REs out there; we should fix borderline cases).
		"""
		quotes = set(['"', "'"])
		def insertLink(matob):
			if matob.group(1) in quotes:
				return matob.group(0)
			return matob.group(1)+'<a href="%s" class="nakedlink">%s</a>'%(
				matob.group(2).replace("&", "&amp;"), matob.group(2))
		return urlRE.sub(insertLink, txt)

	def handleLinks(self, txt):
		txt = self._parseRecursiveElement("[[", "]]", self._formatWikiLink, txt)
		txt = self._parseRecursiveElement("[", "]", self._formatFreeLink, txt)
		return self._formatNakedLinks(txt)

	def handleRedirects(self, txt):
		mat = re.match("(?i)\s*#REDIRECT\s*\[\[(.*)\]\]\s*", txt)
		if mat:
			raise Redirect(self._computeWikiLink(mat.group(1)))
		return txt

	def _buildPage(self, txt):
		refSection = ""
		if self.references:
			refSection = ('<hr width="20%%" align="left"/><ul class="refs">'
				'%s</ul>')%"\n".join(
				['<li><a href="#refback_%03d" name="ref_%03d">[%d]</a> %s</li>'%(
					i+1, i+1, i+1, ref) for i, ref in enumerate(self.references)])
		return txt+refSection

	def toHtml(self, txt, title, naked=False):
		self.hiddenParts = {}
		for processor in self.processors:
			txt = processor(txt)
		txt = self._unhideHidden(
			self._buildPage(txt))
		if naked:
			return txt
		else:
			return jsFunctions+txt


def _testAll(m):
	from smallwp import wpdata
	data = wpdata.makeWpData()
	for title in data.iterPagenames():
		print title.encode("iso-8859-1", "xmlcharrefreplace")
		try:
			m.toHtml(data.getArticle(title), title)
		except wpdata.Error, msg:
			print "***Error: %s"%unicode(msg).encode(
				"iso-8859-1", "xmlcharrefreplace")

def _testSingle(m):
	from smallwp import wpdata
	import os
	if os.path.exists("art.txt"):
		art, title = open("art.txt").read().decode("utf-8"), "Testpage"
	else:
		data = wpdata.makeWpData()
		title = "Japan"
		art = data.getArticle(title)
		open("art.txt", "w").write(art.encode("utf-8"))
	print m.toHtml(art, u"Test").encode("utf-8")

def _testFixed(m):
	print m.toHtml(
	"""{| border="1" cellpadding="3" cellspacing="0" style="border-collapse: collapse;text-align:left;"
{| cellpadding="2" cellspacing="1" class="hintergrundfarbe5" style="float: right; width: 307px; margin: 0 0 0.4em 1em;" summary="Infobox"
! Wappen
! Deutschlandkarte
|- class="hintergrundfarbe2" style="text-align: center;"
| style="width: 50%;" | [[Bild:Erlangen.jpg|140x175px|Wappen  von
	 ]]
| align="center" style="width: 50%;" | border="0" cellspacing="0" cellpadding="0" style="margin:0 0 0 0; border-style:none; border-width:0px; border-collapse:collapse; empty-cells:show"
|<div style="position: relative;"><div style="font-size: 5px; position: absolute; display: block; left:0px; top:0px; padding:0;">[[Bild:Reddot.svg|5px|]]</div>[[Bild:Karte Deutschland.png|140x175px|Deutschlandkarte, Position von  hervorgehoben]]</div>
|}
|-
! colspan="2" | Basisdaten
|- class="hintergrundfarbe2"
| [[Bundesland (Deutschland)|Bundesland]]: ||  [[Bayern]]
		
|- class="hintergrundfarbe2"
| [[Regierungsbezirk]]: || [[Regierungsbezirk Mittelfranken|Mittelfranken]]
|- class="hintergrundfarbe2"
| [[Landkreis]]: || [[Kreisfreie Stadt]]
|- class="hintergrundfarbe2"

|- class="hintergrundfarbe2"

|- class="hintergrundfarbe2"

|- class="hintergrundfarbe2"

|- class="hintergrundfarbe2"

|- class="hintergrundfarbe2"

|- class="hintergrundfarbe2"
| [[Geographische Koordinaten|Koordinaten]]: || <span id="coordinates_3_ObenRechts"><span id="coordinates" class="coordinates" style="white-space:nowrap;">Koordinaten: [http://tools.wikimedia.de/~magnus/geo/geohack.php?language=de&params=49_35_28.24_N_11_01_28.26_E_type: city(0) _region:DE 0°&nbsp;0′&nbsp;N, 0°&nbsp;0′&nbsp;O]</span></span><span class="coordinates">[http://tools.wikimedia.de/~magnus/geo/geohack.php?language=de&params=49_35_28.24_N_11_01_28.26_E_type: city(0) _region:DE 0°&nbsp;0′&nbsp;N, 0°&nbsp;0′&nbsp;O]</span>
|- class="hintergrundfarbe2"
| [[Höhe über dem Meeresspiegel|Höhe]]: || 280&nbsp;m ü.&nbsp;[[Normalnull|NN]]
|- class="hintergrundfarbe2"
| Fläche: || 76.90&nbsp;[[Quadratmeter|km²]]
|- class="hintergrundfarbe2"
| [[Einwohnerzahl|Einwohner]]: || 0 <div class="hintergrundfarbe8">'''Stand bitte im [[Datumsformat|Format]] JJJJ-MM-TT, z.&nbsp;B. 2005-12-31'''</div>
|- class="hintergrundfarbe2"
| [[Bevölkerungsdichte]]: || 0 Einwohner je km²
|- class="hintergrundfarbe2"
| [[Postleitzahl (Deutschland)|Postleitzahl]]en: || 91001–91058 (''alt:''&nbsp;8520)
|- class="hintergrundfarbe2"
| [[Telefonvorwahl|Vorwahl]]en: || 09131, <br />0911 (OT Hüttendorf), <br />09132 (OT Neuses), <br />09135 (OT Dechsendorf)
|- class="hintergrundfarbe2"
| [[Kfz-Kennzeichen (Deutschland)|Kfz-Kennzeichen]]: || ER
|- class="hintergrundfarbe2"
| [[Amtlicher Gemeindeschlüssel|Gemeindeschlüssel]]: || 09 5 62 000
|- class="hintergrundfarbe2" style="display: none;"
|- class="hintergrundfarbe2"
|  
gliederung: || 20 [[Ortsteil]]e
|- class="hintergrundfarbe2"
| Adresse der<br /> 
verwaltung: || Rathausplatz 1<br />91052 Erlangen
|- class="hintergrundfarbe2"
| [[Website|Webpräsenz]]: || <div style="overflow: hidden; width: 148px;">[http://www.erlangen.de/ www.erlangen.de]</div>
|- class="hintergrundfarbe2"
|  [[Oberbürgermeister]] : || [[Siegfried Balleis]]  ([[Christlich-Soziale Union in Bayern|CSU]])
|}
""".decode("utf-8"), "foo").encode("utf-8")

def _test():
	import doctest, wpMarkup
	doctest.testmod(wpMarkup)

if __name__=="__main__":
	m = WikitextFrobnicator("http://foo", None)
	_testFixed(m)
	#_test()
