"""
Code to access the wikipedia data in wikitext
"""

import bz2
import cPickle
import re
import os

from smallwp import wpprefs
from smallwp import wpMarkup
import Looker


class TitleDict(object):
	"""An interface to the index of page titles.

	The main interface to TitleDict is indexing.  An index expression
	either raises PageNotFound or returns a tuple of file name,
	(unpacked) offset and (unpacked) length, which is used by WpData.

	We use a C module looker for binary searching the page index here.
	"""
	def __init__(self, dictFile):
		self.dictFile = dictFile
		self.looker = Looker.Looker(self.dictFile, 0)
	
	def __getitem__(self, key):
		res = self.looker.look((key+"\t").encode("utf-8"))
		if not res.strip():
			# Sadly, wikimedia seems to be somewhat case insenstive
			res = self.looker.look((key.capitalize()+"\t").encode("utf-8"))
			if not res.strip():
				raise wpprefs.PageNotFound("No such page", key)
		items = res.split("\t")
		return items[1], int(items[2]), int(items[3])
	
	def getMatches(self, key, maxMatches=0):
		res = self.looker.look(key.encode("utf-8"), maxMatches).strip()
		return [items.split("\t")[0].decode("utf-8") 
			for items in res.split("\n")]


class SearchResult(object):
	"""is a builder and container for search results.

	It is constructed with a query string, runs the search and keeps
	the result.

	We use the preference item maxTitleMatches to limit the number
	of matches returned (we don't want any desasters with .*).
	"""
	def __init__(self, query, data):
		self.maxLen = wpprefs.getPref("maxTitleMatches")
		self.data = data
		self.results = []
		self.query = query
		self.clipped = False
		self._buildResults()

	def __iter__(self):
		"""returns an iterator for the matching titles.
		"""
		for title in self.results:
			yield title

	def _buildResults(self):
		"""performs the actual search and stores the page titles found
		in this list self.results.
		"""
		try:
			pat = re.compile(self.query, re.I)
		except (KeyError, re.error):
			raise Error("Invalid Query")
		result = []
		for title in self.data.iterPagenames():
			if pat.search(title):
				self.results.append(title)

	def append(self, oneResult):
		"""appends a search result unless we've already seen too much.

		In the latter case, we set the clipped attribute to true.
		"""
		if len(self.results)<self.maxLen:
			self.results.append(oneResult)
		else:
			self.clipped = True

	def getResults(self):
		return self.results

	def getQuery(self):
		return self.query


defaultMeta = {
	"isHTML": False,
	"lang": "en",
}

class WpData(object):
	"""is the interface for accessing the compressed wikipedia data.

	It is constructed with an optional argument specifying the
	location of the wikipedia data.  In this directory, it
	expects the files sections.pickle and pages.index as generated
	by splitstuff (see there) and of course the data itself.
	"""

	def __init__(self, pageDir):
		self.pageDir = pageDir
		self.encoding = "utf-8"
		self._loadMeta()
		self.formatter = wpMarkup.WikitextFrobnicator("/pages/", self)
		try:
			self.pageIndex = TitleDict(self.getTitlesFileName())
			self._prepareSectionInfo()
		except IOError:
			raise Error("Cannot access metadata in directory '%s'."
				"  Did you run splitstuff?"%pageDir)
		self.cachedSearches = {}

	def _loadMeta(self):
		try:
			metaDict = cPickle.load(open(wpprefs.getMetaLocation()))
		except IOError:
			metaDict = _defaultMeta
		self.isHTML = metaDict["isHTML"]
		self.lang = metaDict["lang"]

	def _prepareSectionInfo(self):
		"""unpickles the section info prepared by splitstuff.

		Used by the constructor only.
		"""
		self.sectionInfo = cPickle.load(open(
			os.path.join(self.pageDir, "sections.pickle")))

	def reloadFormatter(self):
		reload(wpMarkup)
		self.formatter = wpMarkup.WikitextFrobnicator("/pages/", self)

	def getArticle(self, pageTitle):
		"""returns the wiki markuped page pageTitle.

		You get back a unicode string, and you're well advised
		to pass in unicode strings as well (though you can get
		away with bytestrings if you care about ascii only).
		"""
		print "Fetching", repr(pageTitle)
		try:
			fName, fPos, size = self.pageIndex[pageTitle]
			srcF = bz2.BZ2File(os.path.join(self.pageDir, fName))
			srcF.seek(fPos)
			return srcF.read(size).decode(self.encoding)
		except (KeyError, IOError), msg:
			# Should probably distinguish these two error sources
			raise wpprefs.PageNotFound(msg, pageTitle)

	def iterPagenames(self):
		"""iterates over all titles known to us.

		We return unicode strings.
		"""
		for line in open(self.getTitlesFileName()):
			yield line.split("\t")[0].decode("utf-8")

	def iterContents(self):
		"""iterates over (title, content) pairs of all articles in the db.

		The items are returned in basically random order.  The building
		of the index takes quite some time.

		Clearly, this is mainly interesting for debugging -- but who knows?
		"""

		def getFileDirectories():
			dir = {}
			for line in open(self.getTitlesFileName()):
				title, fileName, offset, size = line.decode("utf-8").split("\t")
				dir.setdefault(fileName, []).append((int(offset), int(size),
					title))
			for fileDir in dir.values():
				fileDir.sort()
			return dir.items()

		for fName, fileIndex in getFileDirectories():
			srcF = bz2.BZ2File(os.path.join(self.pageDir, fName))
			for _, size, title in fileIndex:
				yield title, srcF.read(size).decode("utf-8")

	def getTitlesFileName(self):
		"""returns the name of the page index for this WpData instance.

		This function is part of the external interface; the only
		piece of data you should use of this file is the titles,
		which is anything before a tab in each line.
		"""
		return os.path.join(self.pageDir, "pages.index")

	def getSections(self):
		return self.sectionInfo

	def getTitlesForSection(self, section):
		"""returns the titles for section.

		Setion has to be a key into sections.  As far as we are
		concerned, this is opaque.
		"""
		return self.pageList[
			self.sections[section][0]:self.sections[section][1]]

	def getSearchResult(self, query):
		"""returns a possibly cached search result.

		TODO: expire old searches.
		"""
		if not self.cachedSearches.has_key(query):
			print "Cache hit:", query
			self.cachedSearches[query] = SearchResult(query, self)
		return self.cachedSearches[query]
	
	def getMatches(self, key, maxMatches=40):
		return self.pageIndex.getMatches(key, maxMatches)

	def has_key(self, key):
		try:
			ignored = self.pageIndex[key]
		except wpprefs.PageNotFound:
			return False
		return True


def makeData(baseDir=wpprefs.getPref("basedir")):
	return WpData(baseDir)


