"""
This module provides preferences and i18n data for the wikipedia server.

Most of this could be put into files.  Since we now have pkg_resources,
that's just a matter of doing it.
"""

import os
import re

from smallwp.fancyconfig import (Configuration, DefaultSection, 
	StringConfigItem, IntConfigItem, BooleanConfigItem, ConfigError,
	makeTxtDocs, readConfiguration)

class Error(Exception):
	"""is our generic Exception.

	Any exception coming out of this package that is not derived
	from this error should be considered a bug of smallwp, unless
	client code did seriously stupid stuff (i.e., breach our interface
	specs).
	"""
	pass


class PageNotFound(Error):
	"""is raised when a page is requested but not found.
	"""
	def __init__(self, msg, title):
		Error.__init__(self, msg)
		self.title = title


class ReportableError(Error):
	"""is an exception containing some user-grokkable message.
	"""


#=======================================================
# Simple settings for the program as a whole
_config = Configuration(
	DefaultSection("Configuration of smallwp",
		StringConfigItem("address", "127.0.0.1", description=
			"Address the server should listen on"),
		IntConfigItem("port", 8780, description=
			"Port the server should listen on"),
		StringConfigItem("sitename", "Small Wikipedia Server",
			description="Name of wikipedia instance for display purposes"),
		IntConfigItem("maxTitleMatches", 1000, description=
			"Maximal number of matches in title search"),
		StringConfigItem("basedir", "/var/share/wikipedia", description=
			"Path to root of wikipedia data"),
		StringConfigItem("mediaBase", 
			"http://upload.wikimedia.org/wikipedia/commons/",
			description="URL fragment to prepend to image links and the like"
				" in HTML dumps"),
		IntConfigItem("mathOutputResolution", 110, description=
			"Resolution (in dpi) of math images"),
		BooleanConfigItem("doTemplates", "False", description=
			"Attempt to render templates? (slow, buggy, but probably worth it)"),
	),
)

def getPref(prefKey):
	return _config.get(prefKey)


def loadConfig():
	try:
		srcName = os.path.join(os.environ["HOME"],
				os.environ.get("SMALLWPCONF", ".smallwp"))
		readConfiguration(_config, None, srcName)
	except ConfigError, msg:
		raise ReportableError(unicode(msg))

loadConfig()

def showHelp():
	print makeTxtDocs(_config)


def getMetaLocation():
	return os.path.join(_config.get("basedir"), "__META__")


#========================================================
# Templates

# The follwoing is a very simple generic template that can be "inherited"
# from to build specific (and rather plain) templates.  Client functions
# need to fill out title and content, templates fill out headMaterial,
# preContent and postContent

_simpleTemplate = """<!DOCTYPE  html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<head>
<title>%%(title)s</title>
<link rel="stylesheet" type="text/css" href="/css/smallwp.css"/>
%(headMaterial)s
<script language="javascript">
function updateList(srcObj) {
	var xmlHttp = new XMLHttpRequest();
	var url = "/titles/"+srcObj.value;
	xmlHttp.open("GET", url, true);
	xmlHttp.onreadystatechange = function() {\
		if (xmlHttp.readyState==4) {
			updatePage("listItems", xmlHttp);
		}
	}
	xmlHttp.send(null);
}

function updatePage(targetId, req) {
	// bad hack -- todo: get rid of innerHTML
	if (req.status==200) {
		respNode = req.responseText;
		targetNode = document.getElementById(targetId);
		targetNode.innerHTML = respNode;
	} else {
		alert("non-200 status: "+string(req.status));
	}
}
//-->
</script>
<link rel="stylesheet" type="text/css" href="/css/mediawiki.css"/>
</head>
<body>
<div id="menu">
<a href="/">Index</a>
<form class="menu" action="/titlesSearch/" method="GET">
<input class="scalable" type="text" name="q" width="20" id="titleFrag" 
	onKeyup="updateList(this)">&nbsp;<input 
	type="submit" value="Title Search"></form>
</form>
<form class="menu" action="/search/" method="GET">
<input class="scalable" type="text" name="q" width="20">&nbsp;<input 
	type="submit" value="RE Search"></form>
</div>
<div id="listItems"></div>
<h1>%%(title)s</h1>
%(preContent)s
%%(content)s
%(postContent)s
</body>
"""


_de_templates = {
# Wikipedia pages
	"page": _simpleTemplate%{
		"headMaterial": "",
		"preContent": "<div class='wptext'>",
		"postContent": "</div>",
	},
# Index pages
	"list": _simpleTemplate%{
		"headMaterial": "",
		"preContent": "",
		"postContent": "",
	},
# Error messages
	"error": _simpleTemplate%{
		"headMaterial": "",
		"preContent": "",
		"postContent": "",
	},
}

_templates = {
	"default": _de_templates,
}

def getTemplate(key):
	return _templates["default"].get(key)


#========================================================
# i18n relevant wikipedia stuff

_wikiLangData = {
	"de": {
		"categoryNamespace": "Kategorie",
		"templateNamespace": "Vorlage",
		"xmlKeptNamespaces": set(["Template", "Vorlage"]),
		"keptNamespaces": set(),
		"htmlIgnoredPat": re.compile("(Vorlage|Diskussion|Bild|Benutzer"
			"|Wikipedia)[~_-]"),
	},
	"ast": {
		"categoryNamespace": u"Categor\xeda",
		"templateNamespace": u"Plant\xeda",
		"xmlKeptNamespaces": set(["Template", "Plantilla", u"Plant\xeda"]),
		"keptNamespaces": set(),
		"htmlIgnoredPat": re.compile("(Plantia|Usuariu|Uiquipedia)[~_-]"),
	},
	"default": {
		"categoryNamespace": "Category",
		"templateNamespace": "Template",
		"keptNamespaces": set(["Template"]),
		"xmlKeptNamespaces": set(),
		"htmlIgnoredPat": re.compile("(Template|User|Image"
			r"|Wikipedia)[~_-]"),
	},
}

def getLangData(lang, key):
	if lang in _wikiLangData:
		return _wikiLangData[lang][key]
	else:
		return _wikiLangData["default"][key]


#========================================================
# Static data for the server
_cssStyle="""#menu * { 
	display: inline;
}

div#menu {
	background:#aaffaa;
	width:100%;
}

.menu input {
	font-size:8pt;
	border: 1px solid black;
}

input.scalable {
	min-width:10%;
	max-width: 30%
}

div.wptext p,li {
	max-width:400px;
}

div.templateBody {
	border: 1pt dotted black;
	background: #cccccc;
	width: 200px;
}

.unresolved {
	color: #cccccc;
}

table.prettytable {
	margin:1em 1em 1em 0; 
	border-style:solid; 
	border-width:1px; 
	border-collapse:collapse; 
	empty-cells:show; 
	caption-side:top;
}

ul.refs {
	list-style-type: none;
}

.refs li {
	padding-top:5px;
}
"""

_staticData = {
	"css": _cssStyle,
}

def getStaticData(key):
	return _staticData[key]
