"""
A little server that provides access to a split and compressed wikipedia
"""

import BaseHTTPServer
import os
import re
import sys
import cgi
import pkg_resources
import urllib
import traceback
from sets import Set as set

import Looker

try:
	from smallwp import htmlmorph
except ImportError:
	sys.stderr.write("BeautifulSoup (or something else required for HTML"
		" processing) not found.\nYou cannot use an HTML dump.\n")
from smallwp import mathimages
from smallwp import wpprefs
from smallwp import wpMarkup
from smallwp import wpdata


class Error(Exception):
	"""is our generic Exception.

	Any exception coming out of this module that is not derived
	from this error should be considered a bug of wpserver, unless
	client code did seriously stupid stuff (i.e., breach our interface
	specs).
	"""
	pass


class NaiveHTMLFormatter:
	"""is a formatter for HTML.

	It doesn't follow python's native formatter interface and is rather
	stupid, but then we don't do fancy stuff here.

	Get your result as an HTML fragment in unicode by calling the
	getResult() method.
	"""
	def __init__(self):
		self.buffer = []
		self.elementStack = []
	
	def getResult(self):
		if self.elementStack:
			raise Error("Element(s) %s still on stack while fomatting"%
				self.elementStack)
		return "".join(self.buffer)

	def addChars(self, chars):
		self.buffer.append(chars)

	def _openElement(self, el, attrs):
		self.elementStack.append(el)
		attString = " ".join(['%s="%s"'%(key, val) 
			for key, val in attrs.iteritems()])
		self.buffer.append("<%s %s>"%(el, attString))

	def closeElement(self):
		self.buffer.append(u"</%s>\n"%self.elementStack.pop())

	def openEnum(self, attrs={}):
		self._openElement("ul", attrs)

	def openListitem(self, attrs={}):
		self._openElement("li", attrs)
	
	def openDiv(self, attrs={}):
		self._openElement("div", attrs)

	def openPre(self, attrs={}):
		self._openElement("pre", attrs)

	def openP(self, attrs={}):
		self._openElement("p", attrs)

	def openAnchor(self, href, attrs={}):
		myAttrs = {"href": href}
		myAttrs.update(attrs)
		self._openElement("a", myAttrs)

	def addFragmentTarget(self, name):
		self._openElement("a", {"name": name})
		self.closeElement()

	def addLink(self, href, anchorText, attrs={}):
		self.openAnchor(href, attrs)
		self.addChars(anchorText)
		self.closeElement()

	def addWikiLink(self, title, specialAnchor=None):
		self.addLink("/pages/%s"%urllib.quote_plus(
			title.encode("utf-8")), specialAnchor or title)


class Translations(object):
	def __init__(self, transFile):
		self.looker = Looker.Looker(transFile, 0)

	def translate(self, path):
		path = path.encode("utf-8")
		res = self.looker.look(path+'\t').decode('utf-8').strip()
		if not res:
			return path
		else:
			return self.translate(res.split('\t')[1])


try:
	_translations = Translations(os.path.join(wpprefs.getPref("basedir"),
		"translations.txt"))
except IOError:
	_translations = None


class WpHandler(BaseHTTPServer.BaseHTTPRequestHandler):
	"""is a handler for the requests to our server.

	The dispatch (i.e. translation from request strings to methods)
	is done through the class attribute methodIndex that maps the 
	first path element of the request to a tuple of method name and
	a set of attributes.  Each method is passed the rest of the query
	string.

	Currently, we only have one method attribute in methodInded: "raw".
	It keeps the dispatcher (_serveRequest) from decoding the rest of
	the request string.  If it is not given, the methods are passed
	decoded unicode strings, otherwise you get urlencoded bytestrings.
	"""
	methodIndex = {
		"pages": ("serveWpPage", []),
		"q": ("serveWpPageStripExt", []),
		"math": ("serveMath", []),
		"opensection": ("doSections", []),
		"titles": ("doTitles", []),
		"titlesSearch": ("doTitleDoc", set(["raw"])),
		"": ("doSections", []),
		"search": ("doSearch", set(["raw"])),
		"css": ("doCss", set(["raw"])),
		"images": ("doImage", []),
	}

	def _serveContent(self, responseCode, contentType,
			content, encoding=None, additionalHeaders={}):
		"""serves any content in a unicode string in UTF8.
		"""
		if encoding:
			content = content.encode(encoding)
		self.send_response(responseCode)
		self.send_header("Content-type", contentType)
		self.send_header("Content-length", "%d"%len(content))
		for key, value in additionalHeaders.iteritems():
			self.send_header(key, value)
		self.end_headers()
		self.wfile.write(content)
		self.wfile.close()

	def serveHtml(self, content):
		"""serves content as HTML.

		Content should be a unicode string.
		"""
		self._serveContent(200, "text/html; charset=utf-8",
			content, "utf-8")

	def _serveError(self, errCode, msg):
		"""serves an error page with an http response code errCode.
		"""
		content = wpprefs.getTemplate("error")%{
				"title": "Error",
				"content": msg,
			}
		self._serveContent(errCode, "text/html; charset=utf-8",
			content, "utf-8")

	def _serveRedirect(self, target):
		"""serves headers that make the browser show target.
		"""
		self._serveContent(301, "text/html; charset=utf-8",
			"", "utf-8", {"Location": target})

	def _serveFromResources(self, path, contentType):
# XXX do if-modified evaluation here
		try:
			self._serveContent(200, contentType, pkg_resources.resource_stream(
				'smallwp', os.path.join('resources', path)).read())
		except IOError:
			self._serveError(404, "No such resource: %s"%path)

	def doImage(self, args):
		"""serves a placeholder image.
		"""
		self._serveFromResources('missing.png', "image/png")

	def doCss(self, path):
		"""serves some css data.

		Our private css still is in wpprefs and is retrieved from there,
		all other css is taken from the pkg_resources.
		"""
		if path=="smallwp.css":
			self._serveContent(200, "text/css",
				wpprefs.getStaticData("css"), "utf-8")
		else:
			self._serveFromResources(path, "text/css")

	def serveMath(self, args):
		pageKey = args[0]
		try:
			image = mathimages.getImage(pageKey)
			self._serveContent(200, "image/png", image, None)
		except mathimages.Error, msg:
			self._serveError(404, str(msg))

	def serveWpPage(self, args):
		"""serves the wiki page with the title given in args[0]
		"""
		pageTitle = args[0]
		wikiData = self.server.data
		wikiPage = wikiData.getArticle(pageTitle)
		try:
			if os.path.exists("art.txt"):
				open("art.txt", "w").write(wikiPage.encode("utf-8"))
		except IOError:
			pass
		if wikiData.isHTML:
			pageText = htmlmorph.morphPage(wikiPage)
		else:
			pageText = wikiData.formatter.toHtml(wikiPage, pageTitle)
		self.serveHtml(
			wpprefs.getTemplate("page")%{
				"title": pageTitle,
				"content": pageText,
			})

	def serveWpPageStripExt(self, args):
		return self.serveWpPage([args[0][:-5]]+args[1:])

	def doSearch(self, encodedForm):
		"""searches for a regular expression in page titles.

		This is somewhat slow since it's a linear search, but I hope
		people don't mind.
		"""
		queryPat = urllib.unquote_plus(cgi.parse_qs(
			encodedForm[1:])["q"][0])
		matches = self.server.data.getSearchResult(queryPat)
		formatter = NaiveHTMLFormatter()

		formatter.openEnum()
		for pageTitle in matches:
			formatter.openListitem()
			formatter.addWikiLink(pageTitle)
			formatter.closeElement()
		formatter.closeElement()

		self.serveHtml(wpprefs.getTemplate("list")%{
			"title":  "Matches for %s"%queryPat,
			"content": formatter.getResult(),
		})
	
	def doSections(self, foldIndices):
		"""handles requests for the section index.

		foldIndices is a list of strings of
		decimal integer representations for the already opened
		sections -- 1/0/10 means the 10th subsubsection of the
		zeroeth subsection of the first section.
		"""
		def renderSectionContents(section, formatter):
			"""generates HTML for rendering the titles within section.

			As generated by splitstuff, section is a tuple of
			(starttitle, startoffset), (endtitle, endoffset), anything.
			The offsets go into the page index file, which is what we
			exploit here.
			"""
			formatter.openEnum()
			f = open(self.server.data.getTitlesFileName())
			f.seek(section[0][1])
			while f.tell()<section[1][1]:
				title = f.readline().split("\t")[0].decode("utf-8")
				formatter.openListitem()
				formatter.addWikiLink(title)
				formatter.closeElement()
			f.close()
			formatter.closeElement()

		def renderSections(sections, formatter, foldIndices, baseUrl):
			"""generates HTML for rendering sections opened along 
			foldIndices.

			Since the sections are a mildly recursive data structure,
			this is recursve, too, breaking off when we either (a)
			have reached a terminal section (i.e., its contents
			are titles) or (b) we have exhausted the list foldIndices
			(i.e., we have opened all folds requested by the user).

			baseUrl contains slash-separated decimal integer indices
			of the fold path up to here -- see the recursive call
			below.
			"""
			if not foldIndices:
				formatter.addFragmentTarget("curFold")
			if not sections[2]:
				renderSectionContents(sections, formatter)
				return
			formatter.openEnum()
			for sectIndex, section in enumerate(sections[2]):
				formatter.openListitem()
				formatter.addLink("%s/%d#curFold"%(baseUrl, sectIndex),
					'%s ... %s'%(section[0][0], section[1][0]))
				if foldIndices and foldIndices[0]==sectIndex:
					renderSections(section, formatter, foldIndices[1:],
						"%s/%d"%(baseUrl, sectIndex))
				formatter.closeElement()
			formatter.closeElement()

		formatter = NaiveHTMLFormatter()
		foldIndices = [int(s) for s in foldIndices if s.strip()]
		renderSections((None, None, self.server.data.getSections()), 
			formatter, foldIndices, "/opensection")
		self.serveHtml(wpprefs.getTemplate("list")%{
			"title": "Index",
			"content": formatter.getResult()})
	
	def _getTitles(self, matchString, maxMatches=40):
		matches = self.server.data.getMatches(matchString, maxMatches)
		formatter = NaiveHTMLFormatter()
		formatter.openP()
		if matches and matches[0]:
			for match in matches:
				formatter.addWikiLink(match)
		else:
			formatter.addChars("No matches for %s -- note that"
				" matching is case sensitive."%matchString)
		formatter.closeElement()
		return formatter.getResult()
	
	def doTitles(self, args):
		matchString = args[0]
		self.serveHtml(self._getTitles(matchString))

	def doTitleDoc(self, encodedForm):
		matchString = urllib.unquote_plus(cgi.parse_qs(
			encodedForm[1:])["q"][0]).decode("utf-8")
		self.serveHtml(wpprefs.getTemplate("list")%{
			"title": "Title search for %s"%matchString,
			"content": self._getTitles(matchString, 400)})

	def _serveRequest(self, func, args):
		"""dispatches requests.

		All request URIs start with one path element ("/bla/"), which is to be
		passed as func.  It is used to dispatch the request
		through methodIndex.  The methods in there receive the
		rest of the request string (args) in a list of unicode strings -- 
		the items are the parts between the slashes.
		"""
		try:
			handler, handlerProps = self.methodIndex[func]
		except KeyError:
			self._serveError(404, "Document with URI '/%s/%s' not found."%(
				func, ("/".join(args))))
			return
		if "raw" in handlerProps:
			args = "/".join(self.path.split("/")[2:])
		getattr(self, handler)(args)

	def splitPath(self, reqPath):
		return [urllib.unquote_plus(s).decode("utf-8") for s in reqPath.split("/")]

	def do_GET(self):
		"""handles GET requests.

		This method is primarily supposed to handle all kinds of
		errors and do something sensible with them.  The real work
		is done by _serveRequest.
		"""
		try:
			parts = self.splitPath(self.path[1:])
			try:
				base = parts[-1]
				if base.endswith(".html") and _translations:
					base = base[:-5]
					newBase = _translations.translate(base).decode("utf-8")
					if newBase.startswith(">"):
						parts = ['pages', newBase[1:]]
				key, args = parts[0], parts[1:]
			except AttributeError:
				self.doSections()
			else:
				self._serveRequest(key, args)
		except wpMarkup.Redirect, msg:
			self._serveRedirect(str(msg))
		except wpprefs.PageNotFound, missingPage:
			self._serveError(404, 
				u"The page %s doesn't exist yet."%(missingPage.title))
		except Error, msg:
			self._serveError(500, unicode(msg))
		except Exception, msg:
			traceback.print_exc()
			self._serveError(500, u"Internal error: '%s'.  Please"
				u" see the server log for a traceback."%unicode(msg))
			
	def log_message(self, format, *args):
		self.server.log_message(format, *args)


class TextLoggingServer(BaseHTTPServer.HTTPServer):
	def __init__(self, data, address, handler):
		self.data = data
		BaseHTTPServer.HTTPServer.__init__(self, address, handler)
		
	def log_message(self, format, *args):
		sys.stderr.write(format%args+"\n")


def runServer():
	"""runs the http server on the address given in preferences.

	This is a no-frills version provided so you can have a text interface.
	This is not what the GUI uses.
	"""
	server_address = (wpprefs.getPref("address"), 
		wpprefs.getPref("port"))
	httpd = TextLoggingServer(wpdata.makeData(), server_address, WpHandler)
	print "Server ready."
	httpd.serve_forever()


def main():
	runServer()
